use super::*;

use ntest::timeout;

use std::sync::mpsc;

fn make_channels(nplayers: usize) -> (Vec<mpsc::Sender<Message>>, Vec<mpsc::Receiver<Message>>, Channels) {
    let mut txs = Vec::new();
    let mut rxs = Vec::new();
    let mut channels = Channels::default();
    for _ in 0..nplayers {
        let (ptx, grx) = mpsc::channel();
        let (gtx, prx) = mpsc::channel();
        txs.push(ptx);
        rxs.push(prx);
        channels.add_channel(gtx, grx);
    }
    (txs, rxs, channels)
}

#[test]
#[timeout(1100)]
fn test_action_challenge() {
    let deck = vec![];
    let discard = vec![];
    let players = vec![
        Player { coins: 2, cards: vec![Duke, Captain] },
        Player { coins: 2, cards: vec![Contessa] },
        Player { coins: 2, cards: vec![Ambassador] },
    ];
    let game = Game {
        deck,
        discard,
        players,
        turn: 0
    };

    // Test non-challengable
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(3);
        assert_eq!(
            game.action_challenge(Move {
                action: Assassinate,
                target: Some(2),
            }, &channels),
            Ok(Phase::Block(Move {
                action: Assassinate,
                target: Some(2),
            }))
        );
    }

    // Test passed challenge
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(3);
        assert_eq!(
            game.action_challenge(Move {
                action: Income,
                target: None,
            }, &channels),
            Ok(Phase::Block(Move {
                action: Income,
                target: None
            }))
        );
    }

    
    // Test failed challenge
    {
        let mut game = game.clone();
        let (txs, rxs, channels) = make_channels(3);
        txs[2].send(Message::Challenge).unwrap();
        txs[2].send(Message::Discard(Ambassador)).unwrap();
        assert_eq!(
            game.action_challenge(Move {
                action: Steal,
                target: Some(2),
            }, &channels),
            Ok(Phase::Block(Move {
                action: Steal,
                target: Some(2)
            }))
        );
        assert_eq!(rxs[2].recv(), Ok(Message::LoseInfluence));
    }

    // Test successful challenge
    {
        let mut game = game.clone();
        let (txs, rxs, channels) = make_channels(3);
        txs[2].send(Message::Challenge).unwrap();
        txs[0].send(Message::Discard(Duke)).unwrap();
        assert_eq!(
            game.action_challenge(Move {
                action: Assassinate,
                target: Some(2),
            }, &channels),
            Ok(Phase::Done),
        );
        assert_eq!(rxs[0].recv(), Ok(Message::LoseInfluence));
    }
}

#[test]
#[timeout(2100)]
fn test_block() {
    let deck = vec![];
    let discard = vec![];
    let players = vec![
        Player { coins: 2, cards: vec![Duke, Captain] },
        Player { coins: 2, cards: vec![Contessa] },
        Player { coins: 2, cards: vec![Ambassador] },
    ];
    let game = Game {
        deck,
        discard,
        players,
        turn: 0
    };


    // Test unblockable
    {
        let (_txs, _rxs, channels) = make_channels(3);
        let mut game = game.clone();
        assert_eq!(
            game.block(Move {
                action: Income,
                target: None,
            }, &channels),
            Ok(Phase::Resolution(Move {
                action: Income,
                target: None
            }))
        );
    }

    // Test target blocked
    {
        let mut game = game.clone();
        let (txs, _rxs, channels) = make_channels(3);
        txs[2].send(Message::Block(Ambassador)).unwrap();
        assert_eq!(
            game.block(Move {
                action: Steal,
                target: Some(2),
            }, &channels),
            Ok(Phase::BlockChallenge(
                Move {
                    action: Steal,
                    target: Some(2),
                },
                Block {
                    blocker: 2,
                    card: Ambassador,
                },
            )),
        );
    }

    // Test target passed block
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(3);
        assert_eq!(
            game.block(Move {
                action: Steal,
                target: Some(2),
            }, &channels),
            Ok(Phase::Resolution(Move {
                action: Steal,
                target: Some(2),
            }))
        );
    }

    // Test anyone blocked
    {
        let mut game = game.clone();
        let (txs, _rxs, channels) = make_channels(3);
        txs[2].send(Message::Block(Duke)).unwrap();
        assert_eq!(
            game.block(Move {
                action: ForeignAid,
                target: None,
            }, &channels),
            Ok(Phase::BlockChallenge(
                Move {
                    action: ForeignAid,
                    target: None,
                },
                Block {
                    blocker: 2,
                    card: Duke,
                },
            )),
        );
    }

    // Test no one blocked
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(3);
        assert_eq!(
            game.block(Move {
                action: ForeignAid,
                target: None,
            }, &channels),
            Ok(Phase::Resolution(Move {
                action: ForeignAid,
                target: None,
            }))
        );
    }

    // Test bad block card
    {
        let mut game = game.clone();
        let (txs, _rxs, channels) = make_channels(3);
        txs[2].send(Message::Block(Contessa)).unwrap();
        assert!(
            game.block(Move {
                action: ForeignAid,
                target: None,
            }, &channels)
                .is_err()
        );
    }
}

#[test]
#[timeout(1100)]
fn test_block_challenge() {
    let deck = vec![];
    let discard = vec![];
    let players = vec![
        Player { coins: 2, cards: vec![Assassin] },
        Player { coins: 2, cards: vec![Contessa] },
    ];
    let game = Game {
        deck,
        discard,
        players,
        turn: 0
    };

    // Test non-challenge
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(2);
        assert_eq!(
            game.block_challenge(
                Move {
                    action: Assassinate,
                    target: Some(1),
                },
                Block {
                    blocker: 1,
                    card: Contessa,
                },
                &channels
            ),
            Ok(Phase::Done)
        );
        assert!(game.discard.is_empty());
    }

    // Test failed challenge
    {
        let mut game = game.clone();
        let (txs, rxs, channels) = make_channels(2);
        txs[0].send(Message::Challenge).unwrap();
        txs[0].send(Message::Discard(Assassin)).unwrap();
        assert_eq!(
            game.block_challenge(
                Move {
                    action: Assassinate,
                    target: Some(1),
                },
                Block {
                    blocker: 1,
                    card: Contessa,
                },
                &channels
            ),
            Ok(Phase::Done)
        );
        assert_eq!(rxs[0].recv(), Ok(Message::LoseInfluence));
        assert!(!game.discard.is_empty());
    }

    // Test successful challenge
    {
        let mut game = game.clone();
        let (txs, rxs, channels) = make_channels(2);
        txs[0].send(Message::Challenge).unwrap();
        txs[1].send(Message::Discard(Contessa)).unwrap();
        assert_eq!(
            game.block_challenge(
                Move {
                    action: ForeignAid,
                    target: None,
                },
                Block {
                    blocker: 1,
                    card: Duke,
                },
                &channels,
            ),
            Ok(Phase::Resolution(Move {
                action: ForeignAid,
                target: None,
            }))
        );
        assert_eq!(rxs[1].recv(), Ok(Message::LoseInfluence));
        assert!(!game.discard.is_empty());
    }
}

#[test]
#[timeout(100)]
fn test_resolution() {
    let deck = vec![Contessa, Contessa];
    let discard = vec![];
    let players = vec![
        Player { coins: 7, cards: vec![Duke, Assassin] },
        Player { coins: 1, cards: vec![Captain] },
    ];
    let game = Game {
        deck,
        discard,
        players,
        turn: 0,
    };

    // Test income
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(2);
        game.resolution(Move {
            action: Income,
            target: None,
        }, &channels).unwrap();
        assert_eq!(game.players[0].coins, 8);
    }

    // Test foreign aid
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(2);
        game.resolution(Move {
            action: ForeignAid,
            target: None,
        }, &channels).unwrap();
        assert_eq!(game.players[0].coins, 9);
    }

    // Test coup / assassinate
    {
        let mut game = game.clone();
        let (txs, rxs, channels) = make_channels(2);
        txs[1].send(Message::Discard(Captain)).unwrap();
        game.resolution(Move {
            action: Coup,
            target: Some(1),
        }, &channels).unwrap();
        assert!(game.players[1].cards.is_empty());
        assert_eq!(rxs[1].recv(), Ok(Message::LoseInfluence));
        assert_eq!(game.discard, vec![Captain]);
        assert_eq!(game.players[0].coins, 0);
    }

    // Test steal
    {
        let mut game = game.clone();
        let (_txs, _rxs, channels) = make_channels(2);
        game.resolution(Move {
            action: Steal,
            target: Some(1),
        }, &channels).unwrap();
        assert_eq!(game.players[0].coins, 8);
        assert_eq!(game.players[1].coins, 0);
    }

    // Test exchange
    {
        let mut game = game.clone();
        let (txs, _rxs, channels) = make_channels(2);
        txs[0].send(Message::Discard(Duke)).unwrap();
        txs[0].send(Message::Discard(Contessa)).unwrap();
        game.resolution(Move {
            action: Exchange,
            target: Some(1),
        }, &channels).unwrap();
        game.players[0].cards.sort();
        assert_eq!(game.players[0].cards, vec![Assassin, Contessa]);
        game.deck.sort();
        assert_eq!(game.deck, vec![Duke, Contessa]);
        assert!(game.discard.is_empty());
    }
 }
